// LorenzMie.js (c) 2017 Jeppe Revall Frisvad
/** 
 * The following source code implements Lorenz-Mie theory
 * using the formulas presented in the SIGGRAPH 2007 paper
 *
 * Computing the Scattering Properties of Participating
 * Media Using Lorenz-Mie Theory
 *
 * By authors: Jeppe Revall Frisvad
 *             Niels Joergen Christensen
 *             Henrik Wann Jensen
 *
 * Code written by Jeppe Revall Frisvad, 2017.
 * Copyright (c) Jeppe Revall Frisvad 2017
 * 
 * Permission is granted to anyone to use this code as 
 * software for any purpose, including commercial applications.
 * However, the software is provided 'as-is' without any warranty.
 */ 

// Complex number operations
function cadd(a, b)
{
  return vec2(a[0] + b[0], a[1] + b[1]);
}

function csub(a, b)
{
  return vec2(a[0] - b[0], a[1] - b[1]);
}

function cmult(a, b)
{
  return vec2(a[0]*b[0] - a[1]*b[1], a[0]*b[1] + a[1]*b[0]);
}

function cdiv(a, b)
{
  var sqr_norm_b = dot(b, b);
  return vec2(dot(a, b)/sqr_norm_b, (a[1]*b[0] - a[0]*b[1])/sqr_norm_b);
}

// Number of terms to include in infinite sums (truncation)
function terms_to_sum(z)
{
  var size = length(z);
  return Math.ceil(size + 4.3*Math.pow(size, 1.0/3.0) + 1.0);
}

// Downward recurrence for A
function A_all_n(z, M)
{
  var A = new Array(M + 2);
  A[M + 1] = vec2(0.0, 0.0);
  for(var n = M; n >= 0; --n)
  {
    var tmp = cdiv(vec2(n + 1.0, 0.0), z);
    A[n] = csub(tmp, cdiv(vec2(1.0, 0.0), cadd(tmp, A[n + 1])));
  }
  return A;
}

// Upward recurrences for B, psi_zeta, and R
function psi_zeta(n, z, A, old_B, old_psi_zeta)
{
  if(n > 0)
  {
    var n_z = cdiv(vec2(n, 0.0), z);
    var tmp = cmult(csub(n_z, A[n - 1]), csub(n_z, old_B));
    old_psi_zeta = cmult(old_psi_zeta, tmp);
  }
  return old_psi_zeta;
}

function B(n, z, A, old_B, old_psi_zeta)
{
  if(n > 0)
  {
    old_B.psi_zeta = psi_zeta(n, z, A, old_B, old_psi_zeta);
    old_B = cadd(A[n], cdiv(vec2(0.0, 1.0), old_B.psi_zeta));
  }
  return old_B;
}

function R(n, z, A, B_n, old_R)
{
  if(n > 0)
  {
    var n_z = cdiv(vec2(n, 0.0), z);
    var tmp = cdiv(cadd(B_n, n_z), cadd(A[n], n_z));
    old_R = cmult(old_R, tmp);
  }
  return old_R;
}

// Lorenz-Mie coefficients
function LorenzMie_ab(n, size, n_p, n_med, A_p, A_med, old_B, old_R, old_psi_zeta)
{
  var x = vec2(size*n_med[0], size*n_med[1]);
  var B_n = B(n, x, A_med, old_B, old_psi_zeta);  
  var R_n = R(n, x, A_med, B_n, old_R);
  var n_med_A_p = cmult(n_med, A_p[n]);
  var n_p_A_med = cmult(n_p, A_med[n]);
  var n_p_A_p = cmult(n_p, A_p[n]);
  var n_med_A_med = cmult(n_med, A_med[n]);
  var n_p_B_n = cmult(n_p, B_n);
  var n_med_B_n = cmult(n_med, B_n);
  var a = cmult(R_n, cdiv(csub(n_med_A_p, n_p_A_med), csub(n_med_A_p, n_p_B_n)));
  var b = cmult(R_n, cdiv(csub(n_p_A_p, n_med_A_med), csub(n_p_A_p, n_med_B_n)));
  var ab = vec4(a[0], a[1], b[0], b[1]);
  ab.old_psi_zeta = n > 0 ? old_B.psi_zeta : old_psi_zeta;
  ab.old_B = B_n;
  ab.old_R = R_n;
  return ab;
}

function LorenzMie_coefs(wavelength, radius, n_p, n_med)
{
  var psize = 2.0*Math.PI*radius/wavelength;
  var x = vec2(psize*n_med[0], psize*n_med[1]);
  var y = vec2(psize*n_p[0], psize*n_p[1]);
  var M = terms_to_sum(x);
  var A_med = A_all_n(x, M);
  var A_p = A_all_n(y, M);
  var e_term = Math.exp(2.0*x[1]);
  var old_psi_zeta = vec2(0.5*(1.0 - Math.cos(2.0*x[0])/e_term), -0.5*Math.sin(2.0*x[0])/e_term);
  var old_B = vec2(0.0, 1.0);
  var old_R = vec2(0.5*(1.0 - Math.cos(2.0*x[0])*e_term), 0.5*Math.sin(2.0*x[0])*e_term);        

  var ab = new Float32Array(M*4);
  for(var n = 0; n < M; ++n)
  {
    var ab_n = LorenzMie_ab(n, psize, n_p, n_med, A_p, A_med, old_B, old_R, old_psi_zeta);
    old_psi_zeta = ab_n.old_psi_zeta;
    old_B = ab_n.old_B;
    old_R = ab_n.old_R;
    ab.set([ab_n[0], ab_n[1], ab_n[2], ab_n[3]], 4*n);
  }
  return ab;
}